#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import glob
import numpy as np
from copy import deepcopy
import bindensity as bind
from scipy.interpolate import UnivariateSpline
from scipy import stats
from ..ANTARESS_general.constant_data import c_light
from ..ANTARESS_conversions.ANTARESS_binning import sub_calc_bins,sub_def_bins
from ..ANTARESS_general.utils import stop,np_where1D,dataload_npz,MAIN_multithread,gen_specdopshift,def_edge_tab,check_data


def def_Mstar(gen_dic,data_inst,inst,data_prop,plot_dic,data_dic,coord_dic):
    """**Stellar master definition.**
    
    Defines master spectra for the star
    
     - used exclusively for flux balance corrections
     - either set to the mean or median of the spectra over visits, or to an input spectrum 
     - the module is ran independently and the master saved for each exposure, so that it can be used independently for the global and local flux balance corrections

    Args:
        TBD
    
    Returns:
        None
    
    """ 
    print('   > Calculating stellar masters')   

    #Calculating data
    if (gen_dic['calc_glob_mast']):
        print('         Calculating data')    

        #Common instrument table, defined in input rest frame
        #    - used to define masters (if a single visit is processed this table will match the visit-specific common table)
        data_com_inst = dataload_npz(data_inst['proc_com_data_path']) 
        wav_Mstar = data_com_inst['cen_bins']
        edge_wav_Mstar = data_com_inst['edge_bins']  
        dim_exp_mast = data_inst['dim_exp']
        Mstar_vis_all = np.zeros([gen_dic[inst]['n_visits']]+dim_exp_mast,dtype=float)*np.nan 
        
        #Calculating visit-specific masters 
        #    - the master can either be calculated using a mean, or a median to avoid introducing spurious variations such as cosmics
        #      the absolute flux level of the master does not matter, as it is used to correct for the relative flux balance correction  
        idx_ord_def_vis = np.zeros(gen_dic[inst]['n_visits'],dtype=object) 
        for ivisit,vis in enumerate(data_inst['visit_list']):
            data_vis = data_inst[vis]

            #Doppler shift used for spectra alignment in the star rest frame (see below)
            #    - beware that systemic RV should be defined or similar between visits if a master common to all visits is used
            specdopshift_star_solbar = 1./(gen_specdopshift(coord_dic[inst][vis]['RV_star_stelCDM'])*gen_specdopshift(data_dic['DI']['sysvel'][inst][vis]))
         
            #Exposure selection    
            if (inst in gen_dic['glob_mast_exp']) and (vis in gen_dic['glob_mast_exp'][inst]) and (gen_dic['glob_mast_exp'][inst][vis]!='all'):
                iexp_mast = gen_dic['glob_mast_exp'][inst][vis]
            else:iexp_mast = range(data_vis['n_in_visit'])
            
            #Processing selected exposures
            flux_vis = np.zeros([len(iexp_mast)]+dim_exp_mast,dtype=float)*np.nan 
            cond_def_vis = np.zeros([len(iexp_mast)]+dim_exp_mast,dtype=bool)
            for isub_exp,iexp in enumerate(iexp_mast):  

                #Upload latest processed data
                data_exp = dataload_npz(data_vis['proc_DI_data_paths']+str(iexp))
                
                #Shifting data from the input rest frame (assumed to be the solar or shifted-solar CDM, receiver) to the star rest frame (source) and resampling on common table
                #    - see gen_specdopshift():
                # w_source = w_receiver / (1+ (rv[s/r]/c))
                # w_star = w_starbar/ (1+ (rv[star/starbar]/c))
                #        = w_solbar / ((1+ (rv[star/starbar]/c)) * (1+ (rv[starbar/solbar]/c)))
                #        = w_solbar / ((1+ (rv_kep/c)) * (1+ (rv_sys/c))) 
                edge_bins_rest = data_exp['edge_bins']*specdopshift_star_solbar[iexp]
                for iord in range(data_inst['nord']): 
                    flux_vis[isub_exp][iord] = bind.resampling(edge_wav_Mstar[iord], edge_bins_rest[iord],data_exp['flux'][iord], kind=gen_dic['resamp_mode'])
                cond_def_vis[isub_exp] = ~np.isnan(flux_vis[isub_exp])

            #Indexes where at least one exposure has a defined bin 
            #    - per order for 2D spectra                       
            cond_def_Mstar_vis = (np.sum(cond_def_vis,axis=0)>0) 

            #Visit master calculated over defined bins
            #    - the co-addition is done over defined bins, to leave at nan those bins where no exposure is defined (otherwise np.nansum returns 0 if only nan are summed)
            #    - bins on the edges of the 1D master over all visits might be undefined if nigthly masters were defined on different ranges  
            #    - we neglect possible changes in instrumental calibration over time, tellurics, and color balance since they are not corrected for yet 
            if gen_dic['glob_mast_mode']=='mean':      

                #Calculate mean of spectra
                #    - the use of the mean naturally gives a stronger weight to spectra with larger flux (and SNR), but it might be biased by strong spurious features
                #    - we do not use weights, as in the present case they would compensate in the weighted mean
                #      after normalizing spectra to the same global flux level, the correct weights would be (see weights_bin_prof):
                # wi = 1/ci, with Fi_norm = ci*Fi 
                #      then Fmast = sum( Fi_norm*wi )/sum(wi) = sum( ci*Fi/ci )/sum(1/ci) = sum(Fi)/sum(ci)
                #      since the overal level of the master does not matter, we can thus simply calculate a mean of the unweighted flux
                for iord in range(data_inst['nord']):
                    for ipix in np_where1D(cond_def_Mstar_vis[iord]):Mstar_vis_all[ivisit][iord,ipix] = np.mean(flux_vis[cond_def_vis[:,iord,ipix],iord,ipix])  
                    
            elif gen_dic['glob_mast_mode']=='med':
                
                #Calculate median of spectra                    
                #     - the use of the median allows mitigating the impact of such features (such as cosmic rays), but it requires that the spectra have comparable flux levels
                #       we apply an empirical normalization to account for all possible source of variations between exposures (duration, stellar variability, earth diffusion, ..)
                for iord in range(data_inst['nord']):

                    #Set each order to same flux level in all exposures, set to average over the visit
                    mean_flux_exp = np.nanmean(flux_vis[:,iord],axis=1)
                    flux_vis[:,iord]*=np.mean(mean_flux_exp)/mean_flux_exp[:,None]
 
                    #Calculate median flux
                    for ipix in np_where1D(cond_def_Mstar_vis[iord]):Mstar_vis_all[ivisit][iord,ipix] = np.median(flux_vis[cond_def_vis[:,iord,ipix],iord,ipix])

            #Defined orders for visit master
            #    - orders might be fully undefined if few bins were defined and spectral tables were different between visits
            #    - we do not remove empty orders to keep the same structure as individual visits
            if data_inst['type']=='spec1D': idx_ord_def_vis[ivisit]=[0]
            elif data_inst['type']=='spec2D':idx_ord_def_vis[ivisit] = np_where1D( np.sum(cond_def_Mstar_vis,axis=1,dtype=bool) > 0)

            #Saving visit-specific master
            if (plot_dic['glob_mast']!='') or (gen_dic[inst]['n_visits']>1):
                np.savez_compressed(gen_dic['save_data_dir']+'Corr_data/Global_Master/'+inst+'_'+vis+'_meas',data = {'cen_bins':wav_Mstar,'flux':Mstar_vis_all[ivisit] ,'cond_def':cond_def_Mstar_vis,'proc_DI_data_paths':data_vis['proc_DI_data_paths'],'specdopshift_star_solbar':specdopshift_star_solbar},allow_pickle=True) 

            #-------------------------------------------------------------------------------------------

            #Shifting and resampling master over the table of each exposure
            #    - to minimize biases in the exposure-to-master ratios 
            #    - even if all exposures and their master are defined over a common table, the master must be resampled after being shifted
            data_mast_exp = {}
            for iexp in range(data_vis['n_in_visit']):

                #Upload latest processed data
                data_exp = np.load(data_vis['proc_DI_data_paths']+str(iexp)+'.npz',allow_pickle=True)['data'].item()     

                #Shifting master from the star rest frame to the exposure rest frame (applying the opposite shift as for the master calculation)
                data_mast_exp['flux'] = np.zeros(data_vis['dim_exp'])*np.nan
                edge_bins_rest = edge_wav_Mstar/specdopshift_star_solbar[iexp]   
                
                #Resampling
                for iord in idx_ord_def_vis[ivisit]:data_mast_exp['flux'][iord] = bind.resampling(data_exp['edge_bins'][iord], edge_bins_rest[iord],Mstar_vis_all[ivisit][iord], kind=gen_dic['resamp_mode'])
                data_mast_exp['cond_def'] = ~np.isnan(data_mast_exp['flux'])                      

                #Saving master for current exposure
                np.savez_compressed(gen_dic['save_data_dir']+'Corr_data/Global_Master/'+inst+'_'+vis+'_'+str(iexp),data = data_mast_exp,allow_pickle=True) 

            ### End of exposures 

            #Theoretical reference master
            #    - two columns: wavelength in star rest frame in A, flux density in arbitrary units  
            if (gen_dic['Fbal_vis']=='theo'):

                #Retrieving input spectrum specific to the visit
                wav_Mstar_theo,Mstar_theo=np.loadtxt(glob.glob(gen_dic['Fbal_refFstar'][inst][vis])[0]).T
                edge_wav_Mstar_theo = def_edge_tab(wav_Mstar_theo,dim=0)
                
                #Check for theoretical spectrum to encompass the full visit range
                min_def_wav_Mstar = np.nanmin(edge_wav_Mstar[0:-1][cond_def_Mstar_vis])
                max_def_wav_Mstar = np.nanmax(edge_wav_Mstar[1::][cond_def_Mstar_vis])
                if (edge_wav_Mstar_theo[0]>min_def_wav_Mstar) or (edge_wav_Mstar_theo[-1]<max_def_wav_Mstar):
                    stop('Extend range of definition of theoretical spectrum to cover visit '+vis)
                
                #Resampling on table of current visit-specific master
                Mstar_ref = np.zeros(dim_exp_mast)*np.nan
                for iord in range(data_inst['nord']): 
                    Mstar_ref[iord] = bind.resampling(edge_wav_Mstar[iord],edge_wav_Mstar_theo,Mstar_theo, kind=gen_dic['resamp_mode']) 
         
                #Saving
                np.savez_compressed(gen_dic['save_data_dir']+'Corr_data/Global_Master/'+inst+'_'+vis+'_theo',data = {'edge_bins':edge_wav_Mstar,'cen_bins':wav_Mstar,'flux':Mstar_ref},allow_pickle=True) 
    
        ### End of visits  

        #Instrument reference master
        #    - all masters may not have the same flux balance, and are thus only co-added at bins where all masters are defined to prevent introducing flux distortions
        #    - if required for multi-visit scaling
        if (gen_dic['Fbal_vis']=='meas') and (gen_dic[inst]['n_visits']>1):
            cond_def_Mstar_vis_all = ~np.isnan(Mstar_vis_all)
            Mstar_ref = np.zeros(dim_exp_mast)*np.nan
            cond_def_Mast = np.zeros(dim_exp_mast,dtype=bool)
            for iord in range(data_inst['nord']):  
                cond_def_Mast[iord] = (np.sum(cond_def_Mstar_vis_all[:,iord,:],axis=0)==gen_dic[inst]['n_visits']) 
                if gen_dic['glob_mast_mode']=='mean':  Mstar_ref[iord,cond_def_Mast[iord]] = np.mean(Mstar_vis_all[:,iord,cond_def_Mast[iord]],axis=0)
                elif gen_dic['glob_mast_mode']=='med': Mstar_ref[iord,cond_def_Mast[iord]] = np.median(Mstar_vis_all[:,iord,cond_def_Mast[iord]],axis=0)
            np.savez_compressed(gen_dic['save_data_dir']+'Corr_data/Global_Master/'+inst+'_meas',data = {'edge_bins':edge_wav_Mstar,'cen_bins':wav_Mstar,'flux':Mstar_ref,'cond_def':cond_def_Mast},allow_pickle=True) 

    #Defining paths and checking data were calculated
    else:
        for vis in data_inst['visit_list']:  
            check_data({iexp:gen_dic['save_data_dir']+'Corr_data/Global_Master/'+inst+'_'+vis+'_'+str(iexp) for iexp in range(data_inst[vis]['n_in_visit'])},vis=vis)                    

    return None
  
    






def corr_Fbal(inst,gen_dic,data_inst,plot_dic,data_prop,data_dic):
    r"""**Main flux balance correction routine.**

    Determines and applies flux balance correction.
    
    If spectra have lost their absolute flux balance, they write as
    
    :math:`F_\mathrm{obs}(w,t) = a(w,t) F(w,t)`
    
    where `a(w,t)` represents instrumental systematics, diffusion by Earth atmosphere, etc.  
    These systematics are assumed to vary slowly with wavelength, so that the shape of the spectra is kept locally.  
    Correcting for these low-frequency variations should then allow us to retrieve the correct shape of planetary and stellar lines.
    
    If we neglect variations due to narrow planetary lines, assumed to be diluted when considering the spectra at low-resolution: 
    
    .. math::
       F_\mathrm{obs}(\mathrm{w \, in \, w_\mathrm{bin}},vis,t) &\sim a(w_\mathrm{bin},t) ( F_\mathrm{\star}(w,vis) LC_\mathrm{tr}(w_\mathrm{bin},vis,t) + F_\mathrm{\star}(w,vis) LC_\mathrm{refl}(w_\mathrm{bin},vis,t) + F_\mathrm{p}^\mathrm{therm}(w_\mathrm{bin},vis,t) )  \\
                             &= a(w_\mathrm{bin},t) F_\mathrm{\star}(w,vis) ( LC_\mathrm{tr}(w_\mathrm{bin},vis,t) + LC_\mathrm{refl}(w_\mathrm{bin},vis,t) + F_\mathrm{p}^\mathrm{therm}(w_\mathrm{bin},vis,t)/F_\mathrm{\star}(w,vis) )  \\                      
                             &= a(w_\mathrm{bin},t) F_\mathrm{\star}(w,vis) \delta_\mathrm{p}(w_\mathrm{bin},vis,t)                        
                          
    where we account for all contributions from the planet. 
    
    We define a reference for the unocculted star :math:`F_\mathrm{ref}(w,vis)`, which can either be a theoretical spectrum, or a measured master spectrum. 
    It is thus possible that :math:`F_\mathrm{ref}(w,vis) = C_\mathrm{ref}(w) F_\mathrm{\star}(w,vis)`, where `A` represents the deviation of the theoretical spectrum from the stellar spectrum, or a   
    combination of the a(w,t) from the spectra used to build the master. Note that the unknown scaling of the flux due to the distance to the star is implicitely included in `A`.  
    The same reference can be used for all spectra obtained with a given instrument, but we obtain better results by calculating references specific to each visit.    
    Assuming that :math:`C_\mathrm{ref}` is dominated by low-frequency variations, we have :math:`F_\mathrm{ref}(w_\mathrm{bin},vis) = C_\mathrm{ref}(w_\mathrm{bin}) F_\mathrm{\star}(w_\mathrm{bin},vis)`

    We bin the data to smooth out noise and high-frequency variations, and to sample `a(w,t)` and the planetary continuum.
    Theoretically we would obtain a cleaner estimate of `a` by first calculating the spectral ratio between each exposure and the reference, and then binning the spectral ratio.   
    In practice however the dispersion of the flux in low-SNR regions leads to spurious variations when calculating this ratio on individual pixels.
    We thus first bin the exposure and master spectra, and then calculate the ratio between those binned spectra.
    
    Rather than binning the data in flux units we first scale it back to raw count units, to avoid artificially increasing the uncertainty and dispersion in a given bin.
    This does not bias the measurement of 'a', considering that:  

    .. math::        
        N_\mathrm{obs}(w_\mathrm{bin},t) &= \sum(w_\mathrm{bin},F_\mathrm{obs}(w,t) dt dw/g_\mathrm{cal}(w))   \\
                     &\sim \sum(w_\mathrm{bin},a(w_\mathrm{bin},t) F_\mathrm{\star}(w,v) dt dw/g_\mathrm{cal}(w))  \\
                     &\sim a(w_\mathrm{bin},t) \delta_\mathrm{p}(w_\mathrm{bin},v,t) dt \sum(w_\mathrm{bin},F_\mathrm{\star}(w,v)/g_\mathrm{cal}(w))     \\    
        N_\mathrm{ref}(w_\mathrm{bin},t) &= \sum(w_\mathrm{bin},F_\mathrm{ref}(w,t) dw/g_\mathrm{cal}(w))        \\
                     &= \sum(w_\mathrm{bin},C_\mathrm{ref}(w_\mathrm{bin}) F_\mathrm{\star}(w,v) dw/g_\mathrm{cal}(w))    \\    
                     &= C_\mathrm{ref}(w_\mathrm{bin}) \sum(w_\mathrm{bin},F_\mathrm{\star}(w,v) dw/g_\mathrm{cal}(w))   
                 
    where we scale the master AFTER its calculation in the star rest frame and shift to the exposure rest frame, to prevent introducing biases
    
    We fit a polynomial to the ratio between the binned exposure spectra and the stellar reference, scaled. :math:`P[w_\mathrm{bin}]` is thus an estimate of: 

    .. math::  
        N_\mathrm{obs}(w_\mathrm{bin},t)/N_\mathrm{ref}(w_\mathrm{bin},t) &= a(w_\mathrm{bin},t) \delta_\mathrm{p}(w_\mathrm{bin},v,t) dt \sum(w_\mathrm{bin},F_\mathrm{\star}(w,v)/g_\mathrm{cal}(w)) / C_\mathrm{ref}(w_\mathrm{bin}) \sum(w_\mathrm{bin},F_\mathrm{\star}(w,v) dw/g_\mathrm{cal}(w))  \\ 
                                  &= a(w_\mathrm{bin},t) \delta_\mathrm{p}(w_\mathrm{bin},vis,t) dt / C_\mathrm{ref}(w_\mathrm{bin}) 
                              
    assuming that :math:`C_\mathrm{ref}` is dominated by low-frequency variations, we obtain 

    :math:`P[w_\mathrm{bin}] ~ a(w_\mathrm{bin},t) \delta_\mathrm{p}(w_\mathrm{bin},vis,t)/C_\mathrm{ref}(w_\mathrm{bin})`
    
    We then extrapolate :math:`P[w_\mathrm{bin}]` at all wavelengths w, and correct spectra using P(w), resulting in :
    
    :math:`F_\mathrm{norm}(w,t) = F_\mathrm{obs}(w,t)/P(w) = F_\mathrm{\star}(w,vis) C_\mathrm{ref}(w) = F_\mathrm{ref}(w,vis)`
    
    low-frequency variations of planetary origins are thus removed together with the systematics, and will be re-injected during the light curve scaling
    
    Finally, we reset the spectra to their original absolute flux level. This is done to remain as close as possible to the measured photon count, which is 
    important in particular to their conversion into stellar CCFs, before it is necessary to rescale the spectra to the exact same flux level, which becomes
    necessary when analyzing the RM effect and atmospheric signals
    
    The correction can be fitted to a smaller spectral band than the full spectrum (eg if the pipeline is set to analyze spectral lines in a given region, or to avoid ranges contaminated by spurious features).
    However the range selected for the fit must be large enough to capture low-frequency variations over the spectral ranges selected for analysis in the pipeline
    
    This correction is applied here, rather than simultaneously with the transit rescaling, because it is necessary to correct for cosmics
    
    When a measured master is used and several visits are processed, we apply first a global correction relative to the master of the visit.
    If requested we then apply the intra-order correction, still using the visit master as reference.
    Finally we apply a global correction based on the ratio between the visit master and a reference, measured at lower resolution than with the visit master.
    This is because deriving directly the correction from the ratio between exposure and instrument master introduces local variations (possibly due to the changes in 
    line shape between epochs) that bias the fitted model. The two-step approach above, where the visit-to-reference correction is measured at low-resolution and high SNR, is more stable
     
    Measuring the stellar continuum on the master and individual spectra, to then do their ratio and measure the flux balance, is not a better solution because to capture well enough the 
    continuum profile the rollin pin has to be small enough, in which case it becomes sensitive to the wiggles. They could then be smoothed out in the continuum ratio, or through a fit with a smoothed or low-order 
    polynomial, but then it becomes equivalent to the present solution    

    Args:
        TBD
    
    Returns:
        None
    
    """     
    txt_print = '   > Correcting spectra for'
    if gen_dic['corr_Fbal']:txt_print+=' global'
    if gen_dic['corr_FbalOrd']:txt_print+=' and intra-order'
    txt_print+=' flux balance'
    print(txt_print)
    if (gen_dic['Fbal_vis']=='meas') and (gen_dic[inst]['n_visits']==1):stop('No need to set flux balance to a reference for a single visit and instrument') 
    if (gen_dic['Fbal_vis'] is None):
        if (gen_dic[inst]['n_visits']>1):stop('Flux balance must be set to a reference for multiple visits')
    else:print('         Final scaling to '+{'meas':'measured','theo':'theoretical'}[gen_dic['Fbal_vis']]+' reference')  
    cond_calc = gen_dic['calc_corr_Fbal'] | gen_dic['calc_corr_FbalOrd'] 

    #Calculating data
    if cond_calc:
        print('         Calculating data')    

        #Global correction
        if gen_dic['corr_Fbal']:

            #Indexes of order to be fitted
            iord_fit_Fbal = range(data_inst['nord_ref'])

            #Deviation between current visit and reference
            if gen_dic['Fbal_vis'] is not None:

                #Calibration profile function
                mean_gdet_func = dataload_npz(gen_dic['save_data_dir']+'Processed_data/Calibration/'+inst+'_mean_gdet')['func'] 

                #Set reference to average of measured visit masters
                if gen_dic['Fbal_vis']=='meas':  
                    data_Mast_ref = dataload_npz(gen_dic['save_data_dir']+'Corr_data/Global_Master/'+inst+'_meas')                     
                    cen_wav_Mstar = data_Mast_ref['cen_bins']
                    low_wav_Mstar = data_Mast_ref['edge_bins'][:,0:-1]
                    high_wav_Mstar = data_Mast_ref['edge_bins'][:,1::]    
                    dwav_Mstar = high_wav_Mstar-low_wav_Mstar

        #Intra-order correction
        if gen_dic['corr_FbalOrd']:        
            range_fit=gen_dic['FbalOrd_range_fit']
            
            #Indexes of order to be fitted
            #    - for "order" mode, the condition also defines orders to be corrected
            iord_fit_Fbal_ord = range(data_inst['nord_ref'])    

        #Default options
        for key in ['Fbal_deg','Fbal_smooth','Fbal_deg_vis','Fbal_smooth_vis','Fbal_deg_ord']:
            if (inst not in gen_dic[key]):gen_dic[key][inst]={}            

        #Process each visit
        corr_func_vis = None
        for ivisit,vis in enumerate(deepcopy(data_inst['visit_list'])):
            data_vis=data_inst[vis]
            data_glob={}

            #Saved paths
            proc_DI_data_paths_new = gen_dic['save_data_dir']+'Corr_data/Fbal/'+inst+'_'+vis+'_' 

            #--------------------------------------------------------
            #Global flux balance
            if gen_dic['corr_Fbal']:

                #Default options
                if (vis not in gen_dic['Fbal_deg_vis'][inst]):gen_dic['Fbal_deg_vis'][inst][vis] = 4
                if (vis not in gen_dic['Fbal_smooth_vis'][inst]):gen_dic['Fbal_smooth_vis'][inst][vis] = 1e-4                  

                #Fitted orders
                if (inst in gen_dic['Fbal_ord_fit']) and (vis in gen_dic['Fbal_ord_fit'][inst]) and (len(gen_dic['Fbal_ord_fit'][inst][vis])>0):
                    iord_fit_list = np.intersect1d(iord_fit_Fbal,gen_dic['Fbal_ord_fit'][inst][vis]) 
                else:iord_fit_list = iord_fit_Fbal                
                
                #Fitted ranges
                if (inst not in gen_dic['Fbal_range_fit']) or ((inst in gen_dic['Fbal_range_fit']) & (vis not in gen_dic['Fbal_range_fit'][inst])):range_fit=[]
                else:range_fit=gen_dic['Fbal_range_fit'][inst][vis] 

                #Measuring deviation between current visit and global instrument master
                #    - see details in corrFbal_vis()
                if gen_dic['Fbal_vis'] is not None:   
                    
                    #Set reference to theoretical master
                    if gen_dic['Fbal_vis']=='theo':                     
                        data_Mast_ref = dataload_npz(gen_dic['save_data_dir']+'Corr_data/Global_Master/'+inst+'_'+vis+'_theo')                     
                        cen_wav_Mstar = data_Mast_ref['cen_bins']
                        low_wav_Mstar = data_Mast_ref['edge_bins'][:,0:-1]
                        high_wav_Mstar = data_Mast_ref['edge_bins'][:,1::]    
                        dwav_Mstar = high_wav_Mstar-low_wav_Mstar    
                        
                    #Retrieve measured visit master
                    data_Mast_vis = dataload_npz(gen_dic['save_data_dir']+'Corr_data/Global_Master/'+inst+'_'+vis+'_meas')
                    
                    #Processing visit and reference masters over their commonly defined pixels, within selected ranges
                    #    - the measured reference for an instrument is only defined in pixels where all visit masters are defined
                    #    - the theoretical reference is defined at all wavelengths
                    if gen_dic['Fbal_vis']=='meas':cond_def_Mast = deepcopy(data_Mast_ref['cond_def'])
                    elif gen_dic['Fbal_vis']=='theo':cond_def_Mast = deepcopy(data_Mast_vis['cond_def'])   
                    if (len(range_fit)>0):
                        cond_sel = np.zeros(data_Mast_ref['flux'].shape,dtype=bool)
                        for bd_band_loc in range_fit:cond_sel|=(low_wav_Mstar>bd_band_loc[0]) & (high_wav_Mstar<bd_band_loc[1])
                        cond_def_Mast &= cond_sel                     
             
                    #Mean flux ratio between visit and reference
                    tot_Fr_vis = corrFbal_totFr(data_inst['nord'],cond_def_Mast,data_dic['DI']['scaling_range'],low_wav_Mstar,high_wav_Mstar,dwav_Mstar,data_Mast_vis['flux'],data_Mast_ref['flux'])
               
                    #Switching fitted data to nu space
                    #    - the scaling must be applied uniformely to the two masters so as not introduce biases
                    low_nu_exp = c_light/high_wav_Mstar[:,::-1]
                    high_nu_exp = c_light/low_wav_Mstar[:,::-1]
                    dnu_exp =  high_nu_exp - low_nu_exp   
                    nu_bins_exp = c_light/cen_wav_Mstar[:,::-1]

                    #Processing requested orders
                    #    - shifting from the star (source) to the solar barycenter (receiver) rest frame
                    #      see gen_specdopshift() :
                    # w_receiver = w_source * (1+ rv[s/r]/c))
                    # w_solbar = w_star * (1+ rv[star/starbar]/c))* (1+ rv[starbar/solbar]/c))
                    #      we neglect the stellar keplerian motion
                    iord_fit_list_def = np.intersect1d(iord_fit_list,np_where1D(np.sum(cond_def_Mast,axis=1)))
                    cen_wav_Mstar_solbar = cen_wav_Mstar*gen_specdopshift(data_dic['DI']['sysvel'][inst][vis])             
                    nu_Mstar_binned = np.zeros(0,dtype=float)
                    Mstar_Fr = np.zeros(0,dtype=float)
                    for iord in iord_fit_list_def:  
                        
                        #Scaling masters back to counts over their common defined pixels in order  
                        #    - median calibration profile must be calculated in solar barycentric rest frame with the function, but defined over the same table that is used in the star rest frame for all masters
                        mean_gdet_Mstar_ord = mean_gdet_func[iord](cen_wav_Mstar_solbar[iord,cond_def_Mast[iord]])[::-1]
                        count_Mstar_vis_ord = data_Mast_vis['flux'][iord,cond_def_Mast[iord]][::-1] /mean_gdet_Mstar_ord                      
                        count_Mstar_ref_ord = data_Mast_ref['flux'][iord,cond_def_Mast[iord]][::-1]/mean_gdet_Mstar_ord
                        
                        #Summing over full order
                        nu_Mstar_binned = np.append(nu_Mstar_binned,np.mean(nu_bins_exp[iord,cond_def_Mast[iord]]))
                        dnu_Mstar_ord = dnu_exp[iord,cond_def_Mast[iord]] 
                        totcount_Mstar_ref = np.sum(count_Mstar_ref_ord*dnu_Mstar_ord)
                        totcount_Mstar_vis = np.sum(count_Mstar_vis_ord*dnu_Mstar_ord) 
                        Mstar_Fr = np.append(Mstar_Fr,totcount_Mstar_vis/totcount_Mstar_ref)

                    #Fitted values
                    id_sort=np.argsort(nu_Mstar_binned)
                    cen_bins_fit = nu_Mstar_binned[id_sort]
                    norm_Fr_fit = Mstar_Fr[id_sort]/tot_Fr_vis   
                    n_bins = len(norm_Fr_fit)
                    cond_fit = np.repeat(True,n_bins)            
           
                    #Fit
                    #    - we neglect uncertainties and assume they are equal for all bins
                    if gen_dic['Fbal_mod']=='pol':corr_func_vis = np.poly1d(np.polyfit(cen_bins_fit[cond_fit], norm_Fr_fit[cond_fit], gen_dic['Fbal_deg_vis'][inst][vis] )) 
                    elif gen_dic['Fbal_mod']=='spline':corr_func_vis = UnivariateSpline(cen_bins_fit[cond_fit], norm_Fr_fit[cond_fit],s=gen_dic['Fbal_smooth_vis'][inst][vis] )

                    #Save correction data for plotting purposes
                    if (plot_dic['Fbal_corr_vis']!=''):
                        data_glob.update({'corr_func_vis':corr_func_vis,'Fbal_wav_bin_vis':c_light/nu_Mstar_binned[::-1],'Fbal_T_binned_vis':Mstar_Fr[::-1],'tot_Fr_vis':tot_Fr_vis}) 
                        data_glob['cond_fit_vis'] = np.zeros(n_bins,dtype=bool)
                        data_glob['cond_fit_vis'][id_sort] = cond_fit[::-1]                             

            #--------------------------------------------------------
            #Intra-order flux balance
            if gen_dic['corr_FbalOrd']:

                #Fitted and corrected orders
                if (inst in gen_dic['FbalOrd_ord_fit']) and (vis in gen_dic['FbalOrd_ord_fit'][inst]) and (len(gen_dic['FbalOrd_ord_fit'][inst][vis])>0):
                    iord_fit_list = np.intersect1d(iord_fit_Fbal_ord,gen_dic['FbalOrd_ord_fit'][inst][vis]) 
                else:iord_fit_list = iord_fit_Fbal_ord

            #Default options       
            if (vis not in gen_dic['Fbal_deg'][inst]):gen_dic['Fbal_deg'][inst][vis] = 4
            if (vis not in gen_dic['Fbal_smooth'][inst]):gen_dic['Fbal_smooth'][inst][vis] = 1e-4     
            if (vis not in gen_dic['Fbal_deg_ord'][inst]):gen_dic['Fbal_deg_ord'][inst][vis] = 4                

            #--------------------------------------------------------
            #Processing all exposures    
            iexp_all = range(data_vis['n_in_visit'])
            common_args = (data_vis['proc_DI_data_paths'],inst,vis,gen_dic['save_data_dir'],range_fit,data_vis['dim_exp'],iord_fit_list,gen_dic['Fbal_bin_nu'],data_dic['DI']['scaling_range'],gen_dic['Fbal_mod'],gen_dic['Fbal_deg'][inst][vis],gen_dic['Fbal_smooth'][inst][vis],gen_dic['Fbal_clip'],data_inst['nord'],data_vis['nspec'],gen_dic['Fbal_range_corr'],\
                            plot_dic['Fbal_corr'],plot_dic['sp_raw'],gen_dic['Fbal_binw_ord'],plot_dic['Fbal_corr_ord'],proc_DI_data_paths_new,gen_dic['Fbal_ord_clip'],gen_dic['resamp_mode'],gen_dic['Fbal_deg_ord'][inst][vis],gen_dic['Fbal_expvar'],data_dic[inst][vis]['mean_gdet_DI_data_paths'],corr_func_vis,gen_dic['corr_Fbal'],gen_dic['corr_FbalOrd'])
            if gen_dic['Fbal_nthreads']>1:tot_Fr_all = MAIN_multithread(corrFbal_vis,gen_dic['Fbal_nthreads'],data_vis['n_in_visit'],[iexp_all],common_args,output = True)                           
            else:tot_Fr_all = corrFbal_vis(iexp_all,*common_args)  
            data_vis['proc_DI_data_paths'] = proc_DI_data_paths_new

            #Save global data 
            if gen_dic['corr_Fbal']:data_glob.update({'tot_Fr_all':tot_Fr_all})
            if gen_dic['corr_FbalOrd']:data_glob['Ord'] = {'iord_corr_list':iord_fit_list}
            np.savez_compressed(proc_DI_data_paths_new+'add',data=data_glob,allow_pickle=True)

    #Updating path to processed data and checking it has been calculated
    else:
        for vis in data_inst['visit_list']:  
            data_vis=data_inst[vis] 
            data_vis['proc_DI_data_paths']=gen_dic['save_data_dir']+'Corr_data/Fbal/'+inst+'_'+vis+'_'         
            check_data({'path':data_vis['proc_DI_data_paths']+str(0)},vis=vis)

    return None


def corrFbal_vis(iexp_group,proc_DI_data_paths,inst,vis,save_data_dir,range_fit,dim_exp,iord_fit_list,Fbal_bin_nu,scaling_range,Fbal_mod,Fbal_deg,Fbal_smooth,Fbal_clip,nord,nspec,Fbal_range_corr,\
                 plot_Fbal_corr,plot_sp_raw,Fbal_binw_ord,plot_Fbal_corr_ord,proc_DI_data_paths_new,Fbal_ord_clip,resamp_mode,Fbal_deg_ord,Fbal_expvar,mean_gdet_DI_data_paths,corr_func_vis,corr_Fbal,corr_FbalOrd):
    r"""**Flux balance correction per visit.**    

    Determines and applies flux balance correction in each visit.    
    
    Args:
        TBD
    
    Returns:
        TBD
    
    """       
    
    #Processing each exposure
    tot_Fr_all = np.zeros(len(iexp_group),dtype=float)
    for isub_exp,iexp in enumerate(iexp_group):
        dic_sav={}
            
        #Upload latest processed data and corresponding visit master
        #    - master has been shifted and resampled onto the exposure table in the input rest frame
        data_exp = dataload_npz(proc_DI_data_paths+str(iexp))
        data_mast_exp = dataload_npz(save_data_dir+'Corr_data/Global_Master/'+inst+'_'+vis+'_'+str(iexp))
        low_wav_exp = data_exp['edge_bins'][:,0:-1]
        high_wav_exp = data_exp['edge_bins'][:,1::] 
        dwav_exp =  high_wav_exp - low_wav_exp                
      
        #Processing master and exposure over their common pixels (defined and outside of selected range)
        cond_fit_all = data_exp['cond_def'] & data_mast_exp['cond_def']
        if (len(range_fit)>0):
            cond_sel = np.zeros(dim_exp,dtype=bool)
            for bd_band_loc in range_fit:cond_sel|=(data_exp['edge_bins'][:,0:-1]>bd_band_loc[0]) & (data_exp['edge_bins'][:,1::]<bd_band_loc[1])
            cond_fit_all &= cond_sel                

        #Retrieve calibration profile associated with current exposure in nu space
        #    - defined in the same frame and over the same table as the exposure spectrum
        mean_gdet_exp = dataload_npz(mean_gdet_DI_data_paths[iexp])['mean_gdet'][:,::-1] 

        #--------------------------------------------------------------------------------
        #Correct flux balance over full spectrum
        if corr_Fbal:

            #Mean flux ratio between exposure and reference
            tot_Fr_all[isub_exp] = corrFbal_totFr(nord,cond_fit_all,scaling_range,low_wav_exp,high_wav_exp,dwav_exp,data_exp['flux'],data_mast_exp['flux'])

            #Switching fitted data to nu space and scaling it from flux to count units
            #    - the scaling must be applied uniformely to the exposure and master spectra (ie, after the master has been calculated) so as not introduce biases
            low_nu_exp = c_light/high_wav_exp[:,::-1]
            high_nu_exp = c_light/low_wav_exp[:,::-1]
            dnu_exp =  high_nu_exp - low_nu_exp   
            nu_bins_exp = c_light/data_exp['cen_bins'][:,::-1]
            count_exp = data_exp['flux'][:,::-1]/mean_gdet_exp
            cond_def_exp =  data_exp['cond_def'][:,::-1]
            count_mast_exp = data_mast_exp['flux'][:,::-1]/mean_gdet_exp 

            #Adding progressively bins of current spectrum and master that will be used to fit the color balance correction
            #    - each order of 2D spectra is processed independently, but contributes to the global binned table  
            bin_exp_dic={}
            for key in ['Fr','varFr','Fmast_tot','cen_bins','low_bins','high_bins','idx_ord_bin']:bin_exp_dic[key] = np.zeros(0,dtype=float) 
            iord_fit_list_def = np.intersect1d(iord_fit_list,np_where1D(np.sum(cond_fit_all,axis=1)))                
            for iord in iord_fit_list_def:
                
                #Force binning of full order at bluest wavelengths
                #    - prevent spline from diverging
                if (inst=='ESPRESSO') and (iord in [0,1]):Fbal_bin_nu_loc = 1000.
                else:Fbal_bin_nu_loc = deepcopy(Fbal_bin_nu)

                #Defining bins
                var_ord = data_exp['cov'][iord][0][::-1]/mean_gdet_exp[iord]**2. 
                bin_bd,raw_loc_dic = sub_def_bins(Fbal_bin_nu_loc,np_where1D(cond_fit_all[iord][::-1]),low_nu_exp[iord],high_nu_exp[iord],dnu_exp[iord],nu_bins_exp[iord],count_exp[iord],Mstar_loc=count_mast_exp[iord],var1D_loc=var_ord)
      
                #Process bins
                nfilled_bins=0
                for ibin,(low_bin_loc,high_bin_loc) in enumerate(zip(bin_bd[0:-1],bin_bd[1:])):
                    bin_loc_dic,nfilled_bins = sub_calc_bins(low_bin_loc,high_bin_loc,raw_loc_dic,nfilled_bins,calc_Fr=True)
                    if len(bin_loc_dic)>0:
                        for key in bin_loc_dic:bin_exp_dic[key] = np.append( bin_exp_dic[key] , bin_loc_dic[key])

                #Store index of order to which the bin belongs to
                bin_exp_dic['idx_ord_bin'] = np.append(bin_exp_dic['idx_ord_bin'],np.repeat(iord,nfilled_bins))

            #Fitted values
            #    - we divide by the exposure-to-master flux ratio C(t), so that only the relative flux balance around the mean exposure flux is corrected for 
            #    - in nu space
            id_sort=np.argsort(bin_exp_dic['cen_bins'])
            cen_bins_fit = bin_exp_dic['cen_bins'][id_sort]
            norm_Fr_fit = bin_exp_dic['Fr'][id_sort]/tot_Fr_all[isub_exp]
            norm_varFr_fit = bin_exp_dic['varFr'][id_sort]/tot_Fr_all[isub_exp]**2.
            n_bins = len(norm_Fr_fit)
            cond_fit = np.repeat(True,n_bins)

            #Uncertainty scaling
            #    - we scale errors because large uncertainties in some parts of the spectra (typically in the blue for ground-based spectra) may bias the polynomial fit in these regions. 
            w_Fr_fit = 1./norm_varFr_fit**Fbal_expvar
            w_Fr_fit = w_Fr_fit/np.mean(w_Fr_fit)

            #Remove extreme outliers
            if Fbal_clip:
                med_prop = np.median(norm_Fr_fit)
                res = norm_Fr_fit - med_prop
                disp_est = stats.median_abs_deviation(res)
                cond_fit[(norm_Fr_fit>med_prop + 10.*disp_est) | (norm_Fr_fit<med_prop-10.*disp_est)] = False               

            #Fit
            if Fbal_mod=='pol':     corr_func = np.poly1d(np.polyfit(cen_bins_fit[cond_fit], norm_Fr_fit[cond_fit], Fbal_deg )) 
            elif Fbal_mod=='spline':corr_func = UnivariateSpline(cen_bins_fit[cond_fit], norm_Fr_fit[cond_fit],s=Fbal_smooth     ,w=w_Fr_fit[cond_fit])

            #Successive fits with automatic identification and exclusion of outliers
            if Fbal_clip:
                thresh = 3.
                for it_res in range(3):
                    
                    #Residuals from previous iteration
                    res = norm_Fr_fit - corr_func(cen_bins_fit)
                    
                    #Sigma-clipping
                    disp_est = np.std(res[cond_fit])
                    cond_fit[(res>thresh*disp_est) | (res<-thresh*disp_est)] = False
         
                    #Fit for current iteration
                    if Fbal_mod=='pol':corr_func = np.poly1d(np.polyfit(cen_bins_fit[cond_fit], norm_Fr_fit[cond_fit],Fbal_deg )) 
                    elif Fbal_mod=='spline':corr_func = UnivariateSpline(cen_bins_fit[cond_fit], norm_Fr_fit[cond_fit],s=Fbal_smooth      ,w=w_Fr_fit[cond_fit])

            #Flatten the order matrix into a 1D table
            n_flat = nord*nspec
            nu_bins_flat = np.reshape(nu_bins_exp,n_flat)
            cond_def_flat = np.reshape(cond_def_exp,n_flat)
            pcorr_T_flat = np.zeros(n_flat,dtype=float)*np.nan

            #Calculate correction 
            #    - the correction is defined over the full range of the original, non-reduced spectra (for plotting purposes) but applied to the reduced spectra 
            pcorr_T_flat[cond_def_flat] = corr_func(nu_bins_flat[cond_def_flat])  

            #Return correction to matrix shape and wavelength dimension
            Fbal_T_fit_all= np.reshape(pcorr_T_flat,[nord,nspec])[:,::-1]
            
            #Define correction range
            #    - we do not correct undefined pixels (correction is set to 1)
            #    - the SNR can be too low in the bluest orders for the correction to be fitted correctly, implying that the noise dominates
            # the flux balance effect. Rather than bias the spectra with the poorly defined polynomial correction, one can chose to leave those bins uncorrected when setting 'Fbal_range_corr'
            #      this will create a sharp transition with corrected bins, but it disappear later on when calculating flux ratios   
            cond_corr = deepcopy(data_exp['cond_def'])
            if len(Fbal_range_corr)>0:
                cond_range=False
                for bd_int in Fbal_range_corr:cond_range |= (low_wav_exp>=bd_int[0]) & (high_wav_exp<=bd_int[1])  
                cond_corr &= cond_range

            #Applying correction
            iord2corr = np_where1D( np.sum( cond_corr,axis=1 )>0 )
            for iord in iord2corr:
                data_corr = np.ones(Fbal_T_fit_all[iord].shape,dtype=float)
                data_corr[cond_corr[iord]] = Fbal_T_fit_all[iord][cond_corr[iord]]                
                data_exp['flux'][iord],data_exp['cov'][iord] = bind.mul_array(data_exp['flux'][iord],data_exp['cov'][iord],1./data_corr)

            #Save independently correction data
            dic_sav.update({'corr_func':corr_func,'corr_func_vis':corr_func_vis})                          
            if (plot_Fbal_corr!='') or (plot_sp_raw!=''):
                dic_sav['Fbal_wav_bin_all'] = np.vstack((c_light/bin_exp_dic['high_bins'][::-1],c_light/bin_exp_dic['cen_bins'][::-1],c_light/bin_exp_dic['low_bins'][::-1]))
                if plot_sp_raw!='':
                    dic_sav['idx_ord_bin'] = bin_exp_dic['idx_ord_bin'][::-1]
                    dic_sav['uncorrected_data_path'] = deepcopy(proc_DI_data_paths+str(iexp)) #save path to data before correction 
                if (plot_Fbal_corr!=''):
                    dic_sav['Fbal_T_binned_all'] = bin_exp_dic['Fr'][::-1]
                    dic_sav['cond_fit'] = np.zeros(n_bins,dtype=bool)
                    dic_sav['cond_fit'][id_sort] = cond_fit[::-1]                       

        #--------------------------------------------------------------------------------
        #Correct flux balance in each order of the latest processed data
        #    - this is a relative flux correction
        #    - we define Fcorr(t) = F(t)*TFr/Fr_mod ~ F(t)*TF(t)*Fmast_mod/(F_mod(t)*TFmast) ~ Fmast_mod*TF(t)/TFmast
        #      so that the corrected spectrum has the same balance as the master stellar spectrum, and keep its original mean flux
        #    - here we assume the bins are small enough that count scaling is not necessary
        if corr_FbalOrd:

            #Processing each order
            Fbal_wav_bin_exp = np.zeros(nord,dtype=object)
            Fbal_T_binned_exp = np.zeros(nord,dtype=object)
            Fbal_T_fit_exp = np.zeros(nord,dtype=object)
            tot_Fr_exp = np.zeros(nord,dtype=float) 
            cond_fit_unsort = np.zeros(nord,dtype=object)
            corr_func={}
            iord_fit_list_def = np.intersect1d(iord_fit_list,np_where1D(np.sum(cond_fit_all,axis=1)))            
            for isub_ord,iord in enumerate(iord_fit_list_def):
                cond_def_sp_exp = data_exp['cond_def'][iord]
                data_corr = np.ones(nspec,dtype=float) 

                #Total flux ratio between exposure and reference, over defined pixels
                idx_def_ord = np_where1D(cond_fit_all[iord])
                bin_dic={'tot_Fr' : np.sum(data_exp['flux'][iord][idx_def_ord]*dwav_exp[iord][idx_def_ord] )/np.sum(data_mast_exp['flux'][iord][idx_def_ord]*dwav_exp[iord][idx_def_ord] ) }   

                #Adding progressively bins that will be used to fit the correction
                #    - we keep this approach because orders with deep telluric lines (for which corrected pixels are undefined)
                # have too many undefined ranges after downsampling with bind.resampling, resulting in poorly defined polynomial corrections
                bin_bd,raw_loc_dic = sub_def_bins(Fbal_binw_ord,cond_fit_all[iord],low_wav_exp[iord],high_wav_exp[iord],dwav_exp[iord],data_exp['cen_bins'][iord],data_exp['flux'][iord]/mean_gdet_exp[iord],Mstar_loc=data_mast_exp['flux'][iord]/mean_gdet_exp[iord])
                for key in ['Fr','Fmast_tot','cen_bins','low_bins','high_bins']:bin_dic[key] = np.zeros(0,dtype=float) 
                for ibin,(low_bin_loc,high_bin_loc) in enumerate(zip(bin_bd[0:-1],bin_bd[1:])):
                    bin_loc_dic,_ = sub_calc_bins(low_bin_loc,high_bin_loc,raw_loc_dic,0,calc_Fr=True)
                    if len(bin_loc_dic)>0:
                        for key in bin_loc_dic:bin_dic[key] = np.append( bin_dic[key] , bin_loc_dic[key])  
                id_sort=np.argsort(bin_dic['cen_bins'])
                for key in ['Fr','Fmast_tot','cen_bins','low_bins','high_bins']:bin_dic[key] = bin_dic[key][id_sort]
                conddef_bin = np.ones(len(bin_dic['cen_bins']),dtype=bool)

                #Fitted values
                #    - we divide by the exposure-to-master flux ratio C(t), so that only the relative flux balance around the mean exposure flux is corrected for 
                cen_bins_fit = bin_dic['cen_bins'][conddef_bin]
                norm_Fr_fit = bin_dic['Fr']/bin_dic['tot_Fr']
                n_bins = len(norm_Fr_fit)
                cond_fit = np.repeat(True,n_bins)
                
                #Remove extreme outliers
                med_prop = np.median(norm_Fr_fit)
                res = norm_Fr_fit - med_prop
                disp_est = stats.median_abs_deviation(res)
                cond_fit[(norm_Fr_fit>med_prop + 10.*disp_est) | (norm_Fr_fit<med_prop-10.*disp_est)] = False               

                #Fit
                corr_func[iord] = np.poly1d(np.polyfit(cen_bins_fit[cond_fit], norm_Fr_fit[cond_fit], Fbal_deg_ord )) 

                #Successive fits with automatic identification and exclusion of outliers
                if Fbal_ord_clip:
                    thresh = 3.
                    for it_res in range(3):
                        
                        #Residuals from previous iteration
                        res = norm_Fr_fit - corr_func[iord](cen_bins_fit)
                        
                        #Sigma-clipping
                        disp_est = np.std(res[cond_fit])
                        cond_fit[(res>thresh*disp_est) | (res<-thresh*disp_est)] = False
             
                        #Fit for current iteration
                        corr_func[iord] = np.poly1d(np.polyfit(cen_bins_fit[cond_fit], norm_Fr_fit[cond_fit], Fbal_deg_ord )) 

                #Calculate and apply relative flux correction
                #    - proxy for F(w,t)/Fstar(w) ~ C(t)*Fbal(w)
                #    - the correction is defined over the full range of the spectra (for plotting purposes), even if then applied to a selected range
                data_corr_rel = np.ones(nspec,dtype=float)
                data_corr_rel[cond_def_sp_exp] = corr_func[iord](data_exp['cen_bins'][iord,cond_def_sp_exp])                     
                data_exp['flux'][iord],data_exp['cov'][iord] = bind.mul_array(data_exp['flux'][iord],data_exp['cov'][iord],1./data_corr_rel)
                data_corr*=data_corr_rel

                #------------------------------------------
                
                #Save independently correction data for plotting purposes
                if plot_Fbal_corr_ord!='':
                    Fbal_wav_bin_exp[iord] = np.vstack((bin_dic['low_bins'],bin_dic['cen_bins'],bin_dic['high_bins']))
                    Fbal_T_binned_exp[iord] = bin_dic['Fr']
                    Fbal_T_fit_exp[iord] = data_corr
                    tot_Fr_exp[iord] = bin_dic['tot_Fr'] 
                    cond_fit_unsort[iord] = np.zeros(n_bins,dtype=bool)       
                    cond_fit_unsort[iord][id_sort] = cond_fit

            #Save independently correction data for plotting purposes
            dic_sav['Ord'] = {'corr_func':corr_func}  
            if plot_Fbal_corr_ord!='':  
                dic_sav['Ord']['Fbal_wav_bin_all'] = Fbal_wav_bin_exp
                dic_sav['Ord']['uncorrected_data_path'] = deepcopy(proc_DI_data_paths+str(iexp)) #save path to data before correction 
                dic_sav['Ord']['Fbal_T_binned_all'] = Fbal_T_binned_exp
                dic_sav['Ord']['Fbal_T_fit_all'] = Fbal_T_fit_exp 
                dic_sav['Ord']['tot_Fr_all'] = tot_Fr_exp       
                dic_sav['Ord']['cond_fit'] = cond_fit_unsort
                dic_sav['Ord']['cen_bins_all'] = data_exp['cen_bins'] 

        ### End of order scaling
        
        #--------------------------------------------------------------------------------
        #Scaling to reference
        #    - performed after the intra-order scaling so that the latter is done with respect to the visit master
        if corr_func_vis is not None:
            pcorr_T_flat = np.zeros(n_flat,dtype=float)*np.nan
            pcorr_T_flat[cond_def_flat]= corr_func_vis(nu_bins_flat[cond_def_flat])
            Fbal_T_fit_all= np.reshape(pcorr_T_flat,[nord,nspec])[:,::-1]
            for iord in iord2corr:
                data_corr = np.ones(Fbal_T_fit_all[iord].shape,dtype=float)
                data_corr[cond_corr[iord]] = Fbal_T_fit_all[iord][cond_corr[iord]]                
                data_exp['flux'][iord],data_exp['cov'][iord] = bind.mul_array(data_exp['flux'][iord],data_exp['cov'][iord],1./data_corr)

        ### End of reference scaling

        #--------------------------------------------------------------------------------
        np.savez_compressed(proc_DI_data_paths_new+str(iexp)+'_add',data=dic_sav,allow_pickle=True)

        #Saving modified data and update paths
        np.savez_compressed(proc_DI_data_paths_new+str(iexp),data = data_exp,allow_pickle=True) 

    return tot_Fr_all



def corrFbal_totFr(nord,cond_def_sp_mast,scaling_range,low_wav,high_wav,dwav,flux_sp,flux_mast):
    r"""    

    Calculates the mean flux ratio between a given spectrum and a reference
    
     - over the same range as requested for the broadband flux scaling, neglecting the small shifts due to alignments
    
    Args:
        TBD
    
    Returns:
        tot_Fr_spec (float): mean flux ratio
    
    """     
    mean_Fsp = 0.
    mean_Fmast = 0.
    dwav_all = 0.
    for iord in range(nord):
        cond_def_ord = deepcopy(cond_def_sp_mast[iord])
        if len(scaling_range)>0:
            cond_def_scal=False 
            for bd_int in scaling_range:cond_def_scal |= (low_wav[iord]>=bd_int[0]) & (high_wav[iord]<=bd_int[1])  
            cond_def_ord &= cond_def_scal 
        dwav_all += np.sum(dwav[iord][cond_def_ord])
        mean_Fsp+=np.sum(flux_sp[iord][cond_def_ord]*dwav[iord][cond_def_ord])
        mean_Fmast+=np.sum(flux_mast[iord][cond_def_ord]*dwav[iord][cond_def_ord])    
    mean_Fsp/=dwav_all
    mean_Fmast/=dwav_all
    tot_Fr_spec = mean_Fsp/mean_Fmast    
    return tot_Fr_spec


















def corr_Ftemp(inst,gen_dic,data_inst,plot_dic,data_prop,coord_dic,data_dic):
    r"""**Temporal flux correction.**    

    Args:
        TBD
    
    Returns:
        TBD
    
    """  
    print('   > Correcting spectra for temporal variations')
    
    #Calculating data
    if (gen_dic['calc_corr_Ftemp']):
        print('         Calculating data')     
    
        #Process each visit independently
        for ivisit,vis in enumerate(data_inst['visit_list']):
            data_vis=data_inst[vis]

            #Scaling range
            if (inst not in gen_dic['Ftemp_range_fit']) or ((inst in gen_dic['Ftemp_range_fit']) & (vis not in gen_dic['Ftemp_range_fit'][inst])):range_fit=[]
            else:range_fit=gen_dic['Ftemp_range_fit'][inst][vis]  

            #Upload common spectral table
            #    - if profiles are defined on different tables they are resampled on this one
            #      if they are already defined on a common table, it is this one, which has been kept the same since the beginning of the routine
            edge_bins_com = (np.load(data_vis['proc_com_data_paths']+'.npz',allow_pickle=True)['data'].item())['edge_bins']
            
            #Global fitting range and tables
            #    - we neglect covariance between bins
            flux_all = np.zeros(data_vis['dim_all'],dtype=float)*np.nan
            sig2_all=np.zeros(data_vis['dim_all'],dtype=float)
            cond_def_fit_all  = np.zeros(data_vis['dim_all'],dtype=bool)
            bjd_all = np.zeros(data_vis['n_in_visit'],dtype=float)
            for iexp in range(data_vis['n_in_visit']): 
                bjd_all[iexp] = coord_dic[inst][vis]['bjd'][iexp]
                
                #Latest processed DI data
                #    - if data were kept on independent tables they need to be resampled on a common one to calculate equivalent fluxes
                data_exp = dataload_npz(data_vis['proc_DI_data_paths']+str(iexp))  
                if (not data_vis['comm_sp_tab']):
                    for iord in range(data_inst['nord']): 
                        flux_all[iexp,iord],cov_temp = bind.resampling(edge_bins_com[iord], data_exp['edge_bins'][iord], data_exp['flux'][iord]  , cov = data_exp['cov'][iord], kind=gen_dic['resamp_mode'])                                                                                                               
                        sig2_all[iexp,iord] = cov_temp[0]
                    cond_def_exp = ~np.isnan(flux_all[iexp])   
                else:
                    flux_all[iexp] = data_exp['flux']
                    sig2_all[iexp]  = data_exp['cov'][iord][0]
                    cond_def_exp = data_exp['cond_def']
    
                #Fitting range, accounting for undefined pixels
                cond_def_fit_all[iexp] = cond_def_exp
                if len(range_fit)>0:
                    cond_range=False
                    for bd_int in range_fit:cond_range |= (edge_bins_com[:,0:-1]>=bd_int[0]) & (edge_bins_com[:,1:]<=bd_int[1])   
                    cond_def_fit_all[iexp] &= cond_range
   
            #Fitting pixels common to all exposures
            cond_def_fit_com  = np.all(cond_def_fit_all,axis=0)   
            
            #Total flux in fitting range
            Tflux_all = np.zeros(data_vis['n_in_visit'],dtype=float)
            Tsig_all = np.zeros(data_vis['n_in_visit'],dtype=float)
            dcen_bin_comm = (edge_bins_com[:,1::] - edge_bins_com[:,0:-1])
            for iord in range(data_inst['nord']):    
                Tflux_all += np.sum(flux_all[:,iord,cond_def_fit_com[iord]]*dcen_bin_comm[iord,cond_def_fit_com[iord]],axis=1)
                Tsig_all += np.sum(sig2_all[:,iord,cond_def_fit_com[iord]]*dcen_bin_comm[iord,cond_def_fit_com[iord]]**2.,axis=1)
            Tsig_all=np.sqrt(Tsig_all)

            #Fitted exposures
            iexp_fit = np_where1D(Tflux_all>0)
            if (inst in gen_dic['idx_nin_Ftemp_fit']) and (vis in gen_dic['idx_nin_Ftemp_fit'][inst]) and (len(gen_dic['idx_nin_Ftemp_fit'][inst][vis])>0): 
                iexp_fit = np.delete(iexp_fit,gen_dic['idx_nin_Ftemp_fit'][inst][vis])

            #Breathing correction
            if inst=='HST':
                
                #faire correction du temps par palier, jutilisais tout le temps ca
                #faire fonction de variation de la phase en polynome
                

                stop('Implement breathing correction from fonction_breath_t_LC.pro and correction_STIS_systematics.pro')
          
            #Temporal correction
            else:
    
                #Fit polynomial as a function of time
                #    - the correction is normalized to maintain average flux over the visit
                corr_func = np.poly1d(np.polyfit(bjd_all[iexp_fit], Tflux_all[iexp_fit]/np.mean(Tflux_all[iexp_fit]) , gen_dic['Ftemp_deg'][inst][vis] , w=1./Tsig_all))

                #Calculate correction over all exposures
                Ftemp_T_fit_all = corr_func(bjd_all)

            #Applying correction to original data
            proc_DI_data_paths_new = gen_dic['save_data_dir']+'Corr_data/Ftemp/'+inst+'_'+vis+'_'                 
            for iexp in range(data_vis['n_in_visit']): 
                data_exp = np.load(data_vis['proc_DI_data_paths']+str(iexp)+'.npz',allow_pickle=True)['data'].item()
                for iord in range(data_inst['nord']): 
                    data_exp['flux'][iord],data_exp['cov'][iord] = bind.mul_array(data_exp['flux'][iord],data_exp['cov'][iord],1./np.repeat(Ftemp_T_fit_all[iexp],data_vis['nspec']))
          
                #Saving modified data and update paths
                np.savez_compressed(proc_DI_data_paths_new+str(iexp),data = data_exp,allow_pickle=True) 
            data_vis['proc_DI_data_paths'] = proc_DI_data_paths_new          
            
            #Save independently correction data                         
            if (plot_dic['Ftemp_corr']!=''):
                dic_sav = {'corr_func':corr_func,'iexp_fit':iexp_fit,'bjd_all':bjd_all,'Tflux_all':Tflux_all,'Tsig_all':Tsig_all}                             
                np.savez_compressed(gen_dic['save_data_dir']+'Corr_data/Ftemp/'+inst+'_'+vis+'_add',data=dic_sav,allow_pickle=True)


    #Updating path to processed data and checking it has been calculated
    else:
        for vis in data_inst['visit_list']:  
            data_vis=data_inst[vis]
            data_vis['proc_DI_data_paths']=gen_dic['save_data_dir']+'Corr_data/Ftemp/'+inst+'_'+vis+'_'  
            check_data({'path':data_vis['proc_DI_data_paths']+str(0)},vis=vis)                

    return None                
            
            
            
            
            
    
    
























#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import numpy as np
from scipy.interpolate import interp1d
from copy import deepcopy
import astropy.convolution.convolve as astro_conv
import bindensity as bind
from pysme import sme as SME
from pysme.linelist.vald import ValdFile
from pysme.abund         import Abund
from pysme.synthesize import synthesize_spectrum
import lmfit
from ctypes import CDLL,c_double,c_int,c_void_p,cast,POINTER
import os as os_system
from ..ANTARESS_analysis.ANTARESS_model_prof import pol_cont,dispatch_func_prof,polycoeff_def,calc_polymodu,calc_linevar_coord_grid
from ..ANTARESS_grids.ANTARESS_star_grid import up_model_star,calc_RVrot,calc_CB_RV,get_LD_coeff
from ..ANTARESS_general.utils import closest,np_poly,np_interp,gen_specdopshift,closest_arr,MAIN_multithread,stop,def_edge_tab

def def_Cfunc_prof():
    r"""**C profile calculation**

    Defines the C function and its parameters used in the optimization

    Args:
        TBD
    
    Returns:
        TBD
    
    """ 

    code_dir = os_system.path.dirname(__file__).split('ANTARESS_grids')[0]
    myfunctions = CDLL(code_dir+'/ANTARESS_analysis/C_grid/Gauss_star_grid.so')
    fun_to_use = myfunctions.C_coadd_loc_gauss_prof
    fun_to_free = myfunctions.free_gaussian_line_grid
    fun_to_use.argtypes = [np.ctypeslib.ndpointer(c_double),
                    np.ctypeslib.ndpointer(c_double),
                    np.ctypeslib.ndpointer(c_double),
                    np.ctypeslib.ndpointer(c_double),
                    np.ctypeslib.ndpointer(c_double),
                    c_int,
                    c_int]
    fun_to_use.restype = c_void_p
    fun_to_free.argtypes = [np.ctypeslib.ndpointer(c_double)]
    fun_to_free.restype = None
    
    return fun_to_use,fun_to_free



def custom_DI_prof(param,x,args=None):
    r"""**Disk-integrated profile: model function**

    Calculates custom disk-integrated stellar profile.
    The disk-integrated profile is built upon a discretized grid of the stellar surface to allow accounting for any type of intensity and velocity field.
    This routine is separated from init_custom_DI_prof() so that it can be used in fitting routines with varying parameters.

    Args:
        TBD
    
    Returns:
        TBD
    
    """ 
    
    #--------------------------------------------------------------------------------
    #Updating stellar grid and line profile grid
    #    - only if routine is called in fit mode, otherwise the default pipeline stellar grid and the scaling from init_custom_DI_prof() are used
    #--------------------------------------------------------------------------------
    if args['fit']:    

        #Updating stellar grid
        #    - if stellar grid is different from the default one 
        if args['var_star_grid'] and (args['unquiet_star'] is None):

            #Update variable stellar properties and stellar grid
            up_model_star(args,param)

            #Update broadband scaling of intrinsic profiles into local profiles
            #    - only necessary if stellar grid is updated
            args['Fsurf_grid_spec'] = theo_intr2loc(args['grid_dic'],args['system_prop'],args,args['ncen_bins'],args['grid_dic']['nsub_star'])     
    
        #--------------------------------------------------------------------------------        
        #Updating intrinsic line profiles
        #    - if the line properties or the stellar grid cells to which they are attributed to vary
        if args['var_line'] or args['var_star_grid']:
            init_st_intr_prof(args,args['grid_dic'],param)
            
    #--------------------------------------------------------------------------------
    #Radial velocities of the stellar surface (km/s)
    #    - an offset is allowed to account for the star/input frame velocity when the model is used on raw data 
    #--------------------------------------------------------------------------------
    rv_surf_star_grid = calc_RVrot(args['grid_dic']['x_st_sky'],args['grid_dic']['y_st'],args['star_params']['istar_rad'],param)[0] + param['rv']
    cb_band = calc_CB_RV(get_LD_coeff(args['system_prop']['achrom'],0),args['system_prop']['achrom']['LD'][0],param['c1_CB'], param['c2_CB'], param['c3_CB'],param)
    if np.max(np.abs(cb_band))!=0.:rv_surf_star_grid += np_poly(cb_band)(args['grid_dic']['mu']).flatten()

    #--------------------------------------------------------------------------------        
    #Coadding local line profiles over stellar disk
    #--------------------------------------------------------------------------------
    icell_list = np.arange(args['grid_dic']['nsub_star'])
    
    #Reducing grid to quiet cells
    if args['unquiet_star'] is not None:
        cond_quiet_star = ~args['unquiet_star']
        rv_surf_star_grid=rv_surf_star_grid[cond_quiet_star]
        args['grid_dic']['mu']=args['grid_dic']['mu'][cond_quiet_star]
        args['flux_intr_grid']=args['flux_intr_grid'][cond_quiet_star]
        icell_list=icell_list[cond_quiet_star]
        args['Fsurf_grid_spec']=args['Fsurf_grid_spec'][cond_quiet_star]
        nsub_star=len(icell_list)
    else:
        nsub_star = len(icell_list)
        cond_quiet_star = np.repeat(True,nsub_star)

    #Set up properties for fast line profile grid calculation
    use_OS_grid=False
    use_C_OS_grid=False
    if 'OS_grid' in args and args['OS_grid']:use_OS_grid=True
    if 'C_OS_grid' in args and args['C_OS_grid']:
        use_OS_grid=False
        use_C_OS_grid=True
    
    #Multithreading
    #    - disabled with theoretical profiles, there seems to be an incompatibility with sme
    if (args['nthreads']>1) and (args['mode']!='theo') and ('prof_grid' not in args['unthreaded_op']) :
        simplified_args={}
        for key in ['mode', 'mac_mode', 'input_cell_all', 'func_prof', 'cen_bins', 'nthreads']:simplified_args[key]=args[key]
        flux_DI_sum=MAIN_multithread(coadd_loc_line_prof,args['nthreads'],nsub_star,[rv_surf_star_grid,icell_list,args['Fsurf_grid_spec'],args['flux_intr_grid'],args['grid_dic']['mu']],(param,simplified_args,),output = True)              

    #Direct call
    else:
        if use_OS_grid:
            for pol_par in args['input_cell_all']:args['input_cell_all'][pol_par]=args['input_cell_all'][pol_par][cond_quiet_star]
            flux_DI_sum=coadd_loc_gauss_prof(rv_surf_star_grid,args['Fsurf_grid_spec'],args)
        elif use_C_OS_grid:
            for pol_par in args['input_cell_all']:args['input_cell_all'][pol_par]=args['input_cell_all'][pol_par][cond_quiet_star]
            Fsurf_grid_spec = args['Fsurf_grid_spec'][:, 0]
            flux_DI_sum = use_C_coadd_loc_gauss_prof(rv_surf_star_grid,Fsurf_grid_spec,args)
        else:flux_DI_sum=coadd_loc_line_prof(rv_surf_star_grid,icell_list,args['Fsurf_grid_spec'],args['flux_intr_grid'],args['grid_dic']['mu'],param,args)
    
    #Co-adding profiles
    DI_flux_norm = np.sum(flux_DI_sum,axis=0)

    #Scaling disk-integrated profile to requested continuum
    #    - DI_flux_norm is returned by the function normalized to unity  
    DI_flux_cont = param['cont']*DI_flux_norm

    #Polynomial continuum level
    #    - P(x) = cont*(1 + a1*rv + a2*rv^2 + a3*rv^3 ... )
    #      defined as x = rv or x=w-wref to provide less leverage to the fit 
    cen_bins_ref = args['cen_bins'] - args['cen_bins_polref']
    DI_flux_mod=DI_flux_cont*pol_cont(cen_bins_ref,args,param)    

    return DI_flux_mod,DI_flux_cont,DI_flux_norm




def init_custom_DI_prof(fixed_args,gen_dic,system_prop,system_spot_prop,theo_dic,star_params,param_in):   
    r"""**Disk-integrated profile: grid initialization**

    Initializes stellar and intrinsic profile grids.

    Args:
        TBD
    
    Returns:
        TBD
    
    """ 
    if isinstance(param_in,lmfit.parameter.Parameters):params={par:param_in[par].value for par in param_in}
    else:params=deepcopy(param_in) 

    #Store various properties potentially overwritten in the fitting procedure
    fixed_args['resamp_mode'] = gen_dic['resamp_mode'] 
    fixed_args['system_prop'] = deepcopy(system_prop) 
    fixed_args['star_params'] = deepcopy(star_params)  
    fixed_args['grid_dic'] = deepcopy(theo_dic)
    fixed_args['system_spot_prop'] = deepcopy(system_spot_prop)

    #------------------------------------------------------------------------
    #Identify variable stellar grid
    #    - condition is that one property controlling the stellar grid is variable OR has a value different from the one in theo_dic use to define the default stellar grid 
    #    - if condition is true, the sky-projected stellar grid and corresponding broadband emission are re-calculated at each step of the minimization for the step parameters
    #    - this option cannot be used with chromatic intensity variations
    #    - this option is not required if only veq is varying 
    #------------------------------------------------------------------------
    stargrid_prop = ['veq','alpha_rot','beta_rot','c1_CB','c2_CB','c3_CB','cos_istar','f_GD','beta_GD','Tpole','A_R','ksi_R','A_T','ksi_T','eta_R','eta_T']
    
    #Fit mode
    #    - grid will be updated in custom_DI_prof() if one of the properties vary
    if fixed_args['fit']:
        
        #Update condition
        fixed_args['var_star_grid']=False
        for par in params:
            if ((par in stargrid_prop) and (params[par] != star_params[par])) or ((par in ['LD_u1','LD_u2','LD_u3','LD_u4']) and (params[par] != system_prop['achrom'][par][0])):fixed_args['var_star_grid']=True    
    
    #Forward mode
    #    - only if properties differ from default ones
    else:
        up_grid = False
        for par in params:
            if ((par in stargrid_prop) and (params[par] != star_params[par])) or ((par in ['LD_u1','LD_u2','LD_u3','LD_u4']) and (params[par] != system_prop['achrom'][par][0])):up_grid = True
        if up_grid:up_model_star(fixed_args,params)

    #------------------------------------------------------------------------
    #Intrinsic line profile grid
    #    - in foward mode: profiles are updated here under condition, and are always used to tile the stellar grid
    #      in fit mode: profiles are updated here if they do not vary during the fit
    #------------------------------------------------------------------------
        
    #Line profile variations
    #    - by default profiles are fixed on initialized values 
    fixed_args['var_line'] = False

    #Theoretical profiles 
    if fixed_args['mode']=='theo':
        fixed_args['abund_sp']=[] 

        #No covariance associated with intrinsic profile
        fixed_args['cov_loc_star']=False
    
        #Spectral table    
        #    - theoretical profiles are defined on a common table in the star rest frame                                                      
        fixed_args['cen_bins_intr'] = np.array(theo_dic['sme_grid']['wave'])  
        fixed_args['edge_bins_intr'] = theo_dic['sme_grid']['edge_bins']                                                                    

        #Update profiles in forward mode
        #    - profiles are only updated if abundances differ from default ones, but are in any case attributed to the stellar grid
        if (not fixed_args['fit']):
            for par in params:
                if 'abund' in par:
                    sp_abund = par.split('_')[1]
                    if np.abs(params[par] - theo_dic['sme_grid']['abund'][sp_abund])>1e-6:fixed_args['abund_sp']+=[sp_abund] 
            init_st_intr_prof(fixed_args,fixed_args['grid_dic'],params)        

        #Fit mode
        #    - profiles are updated during the fit if abundances vary
        else:
            cond_update = False
            for par in params:
                if 'abund' in par:
                    sp_abund = par.split('_')[1]
                    fixed_args['abund_sp']+=[sp_abund] 
                    
                    #Abundance varies
                    if param_in[par].vary:fixed_args['var_line'] = True
                    
                    #Abundance is fixed and differs from initialization
                    elif np.abs(params[par] - theo_dic['sme_grid']['abund'][sp_abund])>1e-6:cond_update = True

            #Update profiles and attribute them to stellar grid if they remain fixed, and the profile and grid are the same as initialization
            if cond_update and (not fixed_args['var_star_grid']):init_st_intr_prof(fixed_args,fixed_args['grid_dic'],params)
          
    #----------------------------------
    #Analytical profiles
    elif (fixed_args['mode']=='ana'):

        #No covariance associated with intrinsic profile
        fixed_args['cov_loc_star']=False    

        #Model function
        #    - calculated directly on disk-integrated spectral table
        #    - if requested, we convolve the final model disk-integrated line with the instrumental LSF before comparison with the measured disk-integrated profile
        #    - the mean flux level of intrinsic profiles is assumed to be unity (see function)   
        if type(fixed_args['func_prof_name'])==str:fixed_args['func_prof'] = dispatch_func_prof(fixed_args['func_prof_name'])
        else:fixed_args['func_prof'] = {inst:dispatch_func_prof(fixed_args['func_prof_name'][inst]) for inst in fixed_args['func_prof_name']}

        #Define profiles in forward mode
        if (not fixed_args['fit']):init_st_intr_prof(fixed_args,fixed_args['grid_dic'],params)
     
        #Fit mode
        #    - there is no default grid of analytical profiles
        #      profiles are calculated directly in each cell, but their properties can be pre-calculated either here (if not fitted) or during the fit (if fitted)
        else:           
            for par in params:
                if (('ctrst_ord' in par) or ('FWHM_ord' in par)) and (param_in[par].vary):fixed_args['var_line'] = True
                
            #Define properties and attribute them to stellar grid if the properties and grid remain fixed
            if (not fixed_args['var_line']) and (not fixed_args['var_star_grid']):init_st_intr_prof(fixed_args,fixed_args['grid_dic'],params)

    #----------------------------------
    #Measured intrinsic profile
    elif fixed_args['mode']=='Intrbin':

        #Covariance associated with intrinsic profile
        fixed_args['cov_loc_star']=True
        
        #Spectral table
        #    - binned profiles are defined on a common table in the star rest frame
        fixed_args['edge_bins_intr'] = fixed_args['edge_bins_Intrbin']
        
        #Define profiles in forward mode
        if (not fixed_args['fit']):init_st_intr_prof(fixed_args,fixed_args['grid_dic'],params)
       
        #Fit mode
        #    - the profiles themselves are fixed, and will only be updated if the stellar grid is updated
        else:

            #Attribute profiles to stellar grid if the latter remains fixed
            if not fixed_args['var_star_grid']:init_st_intr_prof(fixed_args,fixed_args['grid_dic'],params)

    #Reference point for polynomial continuum 
    #    - if several visits are processed we assume their tables are roughly similar
    if (fixed_args['type']=='CCF') or fixed_args['spec2rv']:fixed_args['cen_bins_polref'] = 0.
    else:
        if type(fixed_args['ncen_bins'])==int:fixed_args['cen_bins_polref'] = fixed_args['cen_bins'][int(fixed_args['ncen_bins']/2)]  
        else:
            inst_ref = list(fixed_args['ncen_bins'].keys())[0]
            vis_ref = list(fixed_args['ncen_bins'][inst_ref].keys())[0] 
            cen_pix = int(fixed_args['ncen_bins'][inst_ref][vis_ref]/2) 
            fixed_args['cen_bins_polref'] = fixed_args['cen_bins'][inst_ref][vis_ref][0][cen_pix]  

    return fixed_args



def init_custom_DI_par(fixed_args,gen_dic,system_prop,star_params,params,RV_guess_tab):  
    r"""**Disk-integrated profile: parameter initialization**

    Initializes stellar parameters controlling either disk-integrated or local stellar grids. 
    Fit parameters are initialized to default stellar properties.

    Args:
        TBD
    
    Returns:
        TBD
    
    """ 
    #Stellar grid properties
    #    - all stellar properties are initialized to default stellar values
    #      those defined as variable properties through the settings will be overwritten in 'par_formatting'
    for key,vary,bd_min,bd_max in zip(['veq','veq_spots','alpha_rot','alpha_rot_spots','beta_rot','beta_rot_spots','c1_CB','c2_CB','c3_CB','c1_CB_spots','c2_CB_spots','c3_CB_spots','cos_istar','f_GD','beta_GD','Tpole','A_R','ksi_R','A_T','ksi_T','eta_R','eta_T'],
                                      [False,   False,      False,         False,        False,        False,        False, False,  False,   False,         False,          False,   False,    False,  False,   False, False, False, False, False,  False,   False],
                                      [0.,      0.,         None,          None,         None,         None,         None,  None,   None,    None,          None,           None,     -1.,       0.,     0.,      0.,     0. ,   0.,    0. ,   0.,     0. ,      0.],
                                      [1e4,     1e4,        None,          None,         None,         None,         None,  None,   None,    None,          None,           None,      1.,       1.,     1.,      1e5,    1.,   1e5,    1e5,  1e5,    100.,    100.]):
        if key in star_params:params.add_many((key, star_params[key],   vary,    bd_min,bd_max,None))

    #Properties specific to disk-integrated profiles
    if fixed_args['DI_grid']:
        for ideg in range(1,5):params.add_many(('LD_u'+str(ideg),  system_prop['achrom']['LD_u'+str(ideg)][0],              False,    None,None,None))   

    #Line model properties
    params.add_many(('cont',      fixed_args['flux_cont'],                          False,    None,             None,               None))
    params.add_many(('rv',        RV_guess_tab[0],                                  False,    RV_guess_tab[1],  RV_guess_tab[2],    None)) 
    for ideg in range(1,5):params.add_many(('c'+str(ideg)+'_pol',          0.,              False,    None,None,None)) 

    return params    





def init_st_intr_prof(args,grid_dic,param):
    r"""**Intrinsic stellar profiles: initialization**

    Initializes intrinsic profile grid and properties.
    Called upon initialization of the disk-integrated stellar profile grid, or in fit mode if the stellar grid or the intrinsic profile grid vary

    Args:
        TBD
    
    Returns:
        TBD
    
    """ 
    #Theoretical intrinsic profiles
    if (args['mode']=='theo'):

        #Update line profile series
        sme_grid = deepcopy(args['grid_dic']['sme_grid'])
        if (len(args['abund_sp'])>0):
            for sp in args['abund_sp']:sme_grid['abund'][sp]=param['abund_'+sp]
            gen_theo_intr_prof(sme_grid)
            
        #Interpolating profiles defined over mu grid over full stellar mu grid
        args['flux_intr_grid'] = sme_grid['flux_intr_grid'](grid_dic['mu'])
     
    #Analytic intrinsic profile
    #    - coefficients describing surface properties with polynomial variation of chosen coordinate, required to calculate spectral line profile
    elif (args['mode']=='ana'): 
        
        #Set coordinate grid
        grid_dic['linevar_coord_grid'] = calc_linevar_coord_grid(args['coord_line'],grid_dic)          

        #Set line profile properties
        args['flux_intr_grid'] = np.zeros(grid_dic['nsub_star'])*np.nan   
        args['input_cell_all']={}
        args['coeff_line'] = {}
        inst_list = [args['inst']] if ('inst' in args) else list(args['linevar_par'].keys())
        for inst in inst_list:
            vis_list = [args['vis']] if ('vis' in args) else list(args['linevar_par'][inst].keys())
            for vis in vis_list:  
                for par_loc in args['linevar_par'][inst][vis]:     
                    args['coeff_line'][par_loc] = polycoeff_def(param,args['coeff_ord2name'][inst][vis][par_loc])
                    args['input_cell_all'][par_loc] = calc_polymodu(args['pol_mode'],args['coeff_line'][par_loc],grid_dic['linevar_coord_grid']) 
                                        
    #Measured intrinsic profiles
    #    - attributing to each stellar cell the measured profile with closest coordinate
    elif (args['mode']=='Intrbin'):
        ibin_to_cell = closest_arr(args['cen_dim_Intrbin'], grid_dic[args['coord_line']])
        args['flux_intr_grid'] = np.zeros([grid_dic['nsub_star'],len(args['flux_Intrbin'][0])],dtype=float)
        for icell in range(grid_dic['nsub_star']):args['flux_intr_grid'][icell]=args['flux_Intrbin'][ibin_to_cell[icell]] 
            
    return None




def gen_theo_atm(st_atm,star_params):
    r"""**Theoretical stellar atmosphere**

    Initializes SME grid to generate intrinsic stellar profiles 

    Args:
        TBD
    
    Returns:
        TBD
    
    """

    #Atmosphere structure
    sme_grid = SME.SME_Structure()
    
    #Stellar atmosphere model
    #    - we keep the defaults for sme.atmo.method = "grid" and sme.atmo.geom = "PP"
    sme_grid.atmo.source = st_atm['atm_model']+'.sav'
    
    #NLTE departure
    if ('nlte' in st_atm) and (len(st_atm['nlte'])>0):
        for sp_nlte in st_atm['nlte']:sme_grid.nlte.set_nlte(sp_nlte,st_atm['nlte'][sp_nlte]+'.grd')   
    
    #Set nominal model properties
    #    - vsini is set to 0 so that intrinsic profiles are kept aligned
    sme_grid['teff'] = star_params['Tpole']    
    sme_grid['logg'] = star_params['logg']                
    sme_grid['vsini'] = 0.
    sme_grid['vmic'] = 0.            
    sme_grid['vmac'] = 0.            

    #Wavelength and mu grid of the synthetic spectra
    #    - in A, in the star rest frame
    sme_grid['n_wav'] = int((st_atm['wav_max']-st_atm['wav_min'])/st_atm['dwav'])
    sme_grid_wave = np.linspace(st_atm['wav_min'],st_atm['wav_max'],sme_grid['n_wav']) 
    sme_grid['edge_bins'] = def_edge_tab(sme_grid_wave[None,:][None,:])[0,0]
    sme_grid['wave'] = sme_grid_wave
    
    #Add extreme mu values if not set to prevent extrapolation issues
    id_sort = np.argsort(st_atm['mu_grid'])
    st_atm['mu_grid'] = st_atm['mu_grid'][id_sort]
    if st_atm['mu_grid'][0]>0.:st_atm['mu_grid'] = np.append(0.,st_atm['mu_grid'])
    if st_atm['mu_grid'][-1]<1.:st_atm['mu_grid'] = np.append(st_atm['mu_grid'],1.)
    sme_grid['mu_grid'] = st_atm['mu_grid']
    sme_grid['n_mu'] = len(st_atm['mu_grid'])
    
    #Retrieve linelist and limit to range of spectral grid
    sme_grid['linelist'] = ValdFile(st_atm['linelist'])
    wlcent = sme_grid['linelist']['wlcent']
    cond_within = (wlcent>=st_atm['wav_min']-5.) & (wlcent<=st_atm['wav_max']+5.)
    if True not in cond_within:stop('No VALD transitions within requested range')
    sme_grid['linelist'] = sme_grid['linelist'][cond_within]

    #Abundances
    #    - set by default to solar, from Asplund+2009
    #    - specific abundances are defined as A(X) = log10( N(X)/N(H) ) + 12 
    #    - overall metallicity yields A(X) = Anominal(X) + [M/H] for X != H and He   
    sme_grid['abund'] = Abund.solar()
    if ('abund' in st_atm) and (len(st_atm['abund'])>0):
        for sp_abund in st_atm['abund']:sme_grid['abund'][sp_abund]=st_atm['abund'][sp_abund]
    sme_grid['monh'] = st_atm['MovH'] if 'MovH' in st_atm else 0    
    
    #Intrinsic profile grid
    gen_theo_intr_prof(sme_grid)    
    
    return sme_grid


def gen_theo_intr_prof(sme_grid):
    r"""**Theoretical intrinsic stellar profiles**

    Returns grid of theoretical intrinsic profiles  as a function of :math:`\mu`

    Args:
        TBD
    
    Returns:
        TBD
    
    """    
    #Initialize grid of synthetic spectra
    flux_intr_grid = np.zeros([sme_grid['n_mu'],sme_grid['n_wav']],dtype=float)
    
    #Processing reqested mu
    for imu,mu in enumerate(sme_grid['mu_grid']):
        sme_grid['mu'] = [mu]
    
        #Synthetize spectrum
        sme_spec = synthesize_spectrum(sme_grid)
        flux_intr_grid[imu] = sme_spec.synth[0]

    #Interpolator for profile grid
    sme_grid['flux_intr_grid'] = interp1d(sme_grid['mu_grid'],flux_intr_grid,axis=0)
    
    return None






def theo_intr2loc(grid_dic,system_prop,fixed_args,ncen_bins,nsub_star):
    r"""**Intrinsic to local intensity scaling**

    Returns the scaling values or profiles to scale stellar line profiles from intrinsic to local in each cell.
    
    Intrinsic spectra do not necessarily have the same flux as the disk-integrated profiles, but have been set to the same continuum or total flux (see `proc_intr_data()` and `loc_prof_DI_mast()`).
    Now that the intrinsic and disk-integrated profiles are equivalent in terms of flux we have :math:`F_\mathrm{intr}(w,t,vis) \sim F_\mathrm{DI}(w,vis)`.
    This is valid for broadband fluxes or the continuum of CCFs (see `rescale_profiles()`).
    
    The model disk-integrated profile is defined as 
    
    .. math::    
       F_\mathrm{DI,mod}(w,vis) &= (\sum_{\mathrm{all \, stellar \, cells}} F_\mathrm{intr}(w - w_\mathrm{shift},x,vis) \mathrm{LD}(cell,w) dS ) / A  \\
                                & \sim (\sum_{\mathrm{all \, stellar \, cells}} F_\mathrm{DI}(w - w_\mathrm{shift},vis) \mathrm{LD}(cell,w) dS ) / A
                               
    Where the shift accounts for the stellar surface velocity field. If we neglect these shifts (for spectra) or consider the continuum range (for CCF) then

    .. math::    
       F_\mathrm{DI,mod}(w,vis) \sim F_\mathrm{DI}(w,vis) (\sum_{\mathrm{all \, stellar \, cells}} \mathrm{LD}(cell,w) dS ) / A
       
    Where A is a spectrum   

    .. math::     
       A(w) = \sum_{\mathrm{all \, stellar \, cells}}(\mathrm{LD}(cell,w) dS )

    If the line profile is calculated directly within the fit function we store the scaling profiles.

    Args:
        TBD
    
    Returns:
        TBD
    
    """   
    
    #Grid of broadband intensity variations, at the resolution of current grid
    #    - condition is validated if data is spectral and intensity input is chromatic
    if ('spec' in fixed_args['type']) and ('chrom' in system_prop):

        #Grid of broadband flux variations, at the spectral resolution of the input chromatic properties
        Fsurf_grid_band = grid_dic['Fsurf_grid_star_chrom']
        
        #Grid of broadband flux variations, at the spectral resolution of the local profiles
        #    - interpolation over the profile table if the full range of the profile is larger than the scale of chromatic variations, otherwise closest definition point
        Fsurf_grid_spec = np.zeros([nsub_star,ncen_bins])
        if (fixed_args['edge_bins'][-1]-fixed_args['edge_bins'][0]>system_prop['chrom']['med_dw']):
            for icell in range(nsub_star):
                Fsurf_grid_spec[icell,:] = np_interp(fixed_args['cen_bins'],system_prop['chrom']['w'],Fsurf_grid_band[icell],left=Fsurf_grid_band[icell,0],right=Fsurf_grid_band[icell,-1])
        else:
            iband = closest(system_prop['chrom']['w'],np.median(fixed_args['cen_bins']))
            Fsurf_grid_spec = Fsurf_grid_band[:,iband][:,None]
    else:
        Fsurf_grid_spec = np.tile(grid_dic['Fsurf_grid_star_achrom'][:,0],[ncen_bins,1]).T       

    return Fsurf_grid_spec







def calc_loc_line_prof(icell,rv_surf_star,Fsurf_cell_spec,flux_loc_cell,mu_cell,args,param):
    r"""**Local line profile calculation**

    Calculates local profile from a given cell of the stellar disk.

    Args:
        TBD
    
    Returns:
        TBD
    
    """ 

    #Calculation of analytical intrinsic line profile
    #    - model is always calculated in RV space, and later converted back to wavelength space if relevant
    #    - the model is directly calculated over the RV table at its requested position, rather than being pre-calculated and shifted
    if args['mode']=='ana':
        input_cell = {'cont':1. , 'rv':rv_surf_star,'c1_pol' : 0.,'c2_pol' : 0.,'c3_pol' : 0.,'c4_pol' : 0.}
        for pol_par in args['input_cell_all']:
            input_cell[pol_par] = args['input_cell_all'][pol_par][icell]
        flux_intr=args['func_prof'](input_cell,args['cen_bins'] )[0]
        
        #Convolve intrinsic profile with macroturbulence kernel
        if args['mac_mode'] is not None:
    
            #Mactroturbulence kernel table     
            #    - centered spectral table with same pixel widths as the band spectrum the kernel is associated to
            #    - a range of 3.15 times the FWHM already contains 99.98% of a Gaussian LSF integral, we thus use 5 times its value 
            #    - G(rv) = exp(- (rv/ sqrt(2)*sig )^2  ) with  = FWHM/(2 sqrt(2ln(2)))   
            #      with eta = sqrt(2)*sig
            #           sig = FWHM/(2 sqrt(2ln(2)))  
            #      thus FWHM = eta*(2 sqrt(ln(2)))  
            #      we assume a conservative FWHM = 50km/s for macroturbulence
            dbins = args['cen_bins'][1]-args['cen_bins'][0]        
            hnkern=np.int(np.ceil(2.125*50./dbins)+1)
            cen_bins_kernel=dbins*np.arange(-hnkern,hnkern+1)
            
            #Calculate local macroturbulence kernel
            cos_mu_cell = mu_cell
            sin_mu_cell = np.sqrt(1. - mu_cell**2.)
            macro_kern_loc = args['mac_mode_func'](cen_bins_kernel,param,cos_mu_cell,sin_mu_cell)
            macro_kern_loc=macro_kern_loc/np.sum(macro_kern_loc) 
    
            #Convolution
            #    - bins must have the same size in a given table
            flux_intr=astro_conv(flux_intr,macro_kern_loc,boundary='extend')          
        
    #Shift stored intrinsic profile from the photosphere (source) to the star (receiver) rest frame     
    #    - see gen_specdopshift():
    # w_receiver = w_source * (1+ (rv[s/r]/c))
    # w_star = w_photo * (1+ (rv[photo/star]/c))
    #        = w_photo * (1+ (rv_surf/c))
    #    - models are pre-calculated and then shifted, since the line profile and the shift are independent
    elif (args['mode'] in ['theo','Intrbin']):
        if ('spec' in args['type']):edge_bins_surf = args['edge_bins_intr']*gen_specdopshift(rv_surf_star)        
        elif (args['type']=='CCF'):edge_bins_surf = args['edge_bins_intr'] + rv_surf_star
        flux_intr = bind.resampling(args['edge_bins'],edge_bins_surf,flux_loc_cell, kind=args['resamp_mode'])            
    
    #Continuum scaling into local line profiles
    #    - default continuum of intrinsic profiles is set to 1 so that it can be modulated chromatically here
    flux_loc=flux_intr*Fsurf_cell_spec
    
    return flux_loc




def coadd_loc_line_prof(rv_surf_star_grid,icell_list,Fsurf_grid_spec,flux_intr_grid,mu_grid,param,args):
    r"""**Local line profile co-addition**

    Cumulates local profiles from each cell of the stellar disk.

    Args:
        TBD
    
    Returns:
        TBD
    
    """ 
    flux_DI_sum=[]
    for isub,(icell,rv_surf_star,flux_intr_cell,mu_cell) in enumerate(zip(icell_list,rv_surf_star_grid,flux_intr_grid,mu_grid)):
        flux_DI_sum+=[calc_loc_line_prof(icell,rv_surf_star,Fsurf_grid_spec[isub],flux_intr_cell,mu_cell,args,param)]
    return flux_DI_sum




def coadd_loc_gauss_prof(rv_surf_star_grid, Fsurf_grid_spec, args):
    r"""**Local Gaussian line co-addition**

    Oversimplified way of cumulating the local profiles from each cell of the stellar disk. 
    This version assumes gaussian line profiles in each cell.

    Args:
        TBD
    
    Returns:
        TBD
    
    """ 
    #Define necessay grids    
    true_rv_surf_star_grid = np.tile(rv_surf_star_grid, (args['ncen_bins'], 1)).T
    model_table = np.ones((Fsurf_grid_spec.shape[0], args['ncen_bins']), dtype=float) * args['cen_bins']
    cont_grid = np.ones((Fsurf_grid_spec.shape[0], args['ncen_bins']))
    sqrt_log2 = np.sqrt(np.log(2.))
    ctrst_grid = np.tile(args['input_cell_all']['ctrst'], (args['ncen_bins'], 1)).T
    FWHM_grid = np.tile(args['input_cell_all']['FWHM'], (args['ncen_bins'], 1)).T
    
    #Make grid of profiles    
    gaussian_line_grid = cont_grid*(1.-ctrst_grid*np.exp(-(2.*sqrt_log2*(model_table-true_rv_surf_star_grid)/FWHM_grid)**2))

    gaussian_line_grid *= Fsurf_grid_spec

    return gaussian_line_grid



def use_C_coadd_loc_gauss_prof(rv_surf_star_grid, Fsurf_grid_spec, args):
    r"""**C++ local Gaussian line co-addition**

    C++ implementation of `coadd_loc_gauss_prof()`.

    Args:
        TBD
    
    Returns:
        TBD
    
    """ 
    Fsurf_grid_spec_shape0 = len(Fsurf_grid_spec)
    ncen_bins = args['ncen_bins']
    gauss_grid_ptr = args['fun_to_use'](rv_surf_star_grid, args['input_cell_all']['ctrst'], args['input_cell_all']['FWHM'], 
        args['cen_bins'], Fsurf_grid_spec*10, ncen_bins, Fsurf_grid_spec_shape0)
    gauss_grid = np.frombuffer(cast(gauss_grid_ptr, POINTER(c_double * (ncen_bins * Fsurf_grid_spec_shape0))).contents, dtype=np.float64)
    truegauss_grid = gauss_grid.reshape((Fsurf_grid_spec_shape0, ncen_bins))/10
    args['fun_to_free'](gauss_grid)
    return truegauss_grid






